from contextlib import nullcontext as does_not_raise
from math import pi

import einops
import imageio.v3 as iio
import numpy as np
import pytest
import scipy.ndimage
import torch
from numpy.random import randint
from skimage import color

import plenoptic as po
from conftest import DEVICE, IMG_DIR
from plenoptic.data.fetch import fetch_data
from plenoptic.tools.data import _check_tensor_equality


# used for load_images test as a folder that contains no images
@pytest.fixture()
def folder_with_no_images():
    return fetch_data("portilla_simoncelli_test_vectors.tar.gz")


class TestData:
    def test_load_images_fail(self):
        with pytest.raises(ValueError, match="All images must be the same shape"):
            po.load_images(
                [
                    IMG_DIR / "256" / "einstein.pgm",
                    IMG_DIR / "mixed" / "bubbles.png",
                ]
            )

    def test_load_images_non_image(self, folder_with_no_images):
        err = pytest.raises(ValueError, match="None of the files found")
        warn = pytest.warns(UserWarning, match="Unable to load in file")
        with err, warn:
            po.load_images(folder_with_no_images.parent)

    def test_load_images_sort(self):
        imgs = po.load_images(IMG_DIR / "256")
        sorted_paths = [
            "color_wheel.jpg",
            "curie.pgm",
            "einstein.pgm",
            "metal.pgm",
            "nuts.pgm",
        ]
        sorted_paths = [IMG_DIR / "256" / f for f in sorted_paths]
        imgs_2 = []
        for p in sorted_paths:
            img = iio.imread(p)
            img = img / 255
            if img.ndim == 3:
                img = color.rgb2gray(img)
            imgs_2.append(imgs)
        imgs_2 = torch.as_tensor(np.array(imgs_2), dtype=torch.float32)
        torch.equal(imgs, imgs_2)

    def test_load_images_paths(self):
        sorted_paths = [
            "color_wheel.jpg",
            "curie.pgm",
            "einstein.pgm",
            "metal.pgm",
            "nuts.pgm",
        ]
        sorted_paths = [IMG_DIR / "256" / f for f in sorted_paths]
        imgs = po.load_images(sorted_paths)
        imgs_2 = []
        for p in sorted_paths:
            img = iio.imread(p)
            img = img / 255
            if img.ndim == 3:
                img = color.rgb2gray(img)
            imgs_2.append(imgs)
        imgs_2 = torch.as_tensor(np.array(imgs_2), dtype=torch.float32)
        torch.equal(imgs, imgs_2)

    def test_load_images_custom_sort(self):
        imgs = po.load_images(IMG_DIR / "256", sorted_key=lambda x: x.name[1])
        sorted_paths = [
            "metal.pgm",
            "einstein.pgm",
            "color_wheel.jpg",
            "curie.pgm",
            "nuts.pgm",
        ]
        sorted_paths = [IMG_DIR / "256" / f for f in sorted_paths]
        imgs_2 = []
        for p in sorted_paths:
            img = iio.imread(p)
            img = img / 255
            if img.ndim == 3:
                img = color.rgb2gray(img)
            imgs_2.append(imgs)
        imgs_2 = torch.as_tensor(np.array(imgs_2), dtype=torch.float32)
        torch.equal(imgs, imgs_2)

    def test_load_images_custom_sort_fail(self):
        imgs = list((IMG_DIR / "256").iterdir())
        with pytest.raises(ValueError, match="When paths argument is"):
            imgs = po.load_images(imgs, sorted_key=lambda x: x.name[1])

    # these two warnings are triggered during a call to pkg_resources from imageio. The
    # first one only happens if either sphinxcontrib-apidoc or sphinxcontrib-jsmath is
    # also in your environment (they will be if your environment includes sphinx, which
    # also gets installed by numpydoc). hopefully, should be fixed:
    # https://github.com/imageio/imageio/issues/1137
    @pytest.mark.filterwarnings(
        "ignore:Deprecated call to `pkg_resources:DeprecationWarning"
    )
    @pytest.mark.filterwarnings(
        "ignore:pkg_resources is deprecated as an API:UserWarning"
    )
    def test_load_images_some_non_image(self):
        test_dir = fetch_data("load_image_test.tar.gz")
        warn = pytest.warns(UserWarning, match="Unable to load in file")
        with warn:
            po.load_images(test_dir)

    def test_load_image_notfound(self, tmp_path):
        with pytest.raises(FileNotFoundError, match="File .* not found!"):
            po.load_images(tmp_path / "test.png")

    def test_load_images_notfound(self, tmp_path, einstein_img):
        iio.imwrite(tmp_path / "einstein.pgm", po.to_numpy(einstein_img).squeeze())
        with pytest.raises(FileNotFoundError, match="File .* not found!"):
            po.load_images([tmp_path / "test.png", tmp_path / "einstein.pgm"])

    @pytest.mark.parametrize("filename", ["color_wheel.jpg", "einstein.pgm"])
    def test_load_images_color(self, filename):
        img = po.load_images(IMG_DIR / "256" / filename, as_gray=False)
        assert img.shape[1] == 3, "Didn't load image in color!"


class TestSignal:
    def test_polar_amplitude_zero(self):
        a = torch.rand(10, device=DEVICE) * -1
        b = po.tools.rescale(torch.randn(10, device=DEVICE), -pi / 2, pi / 2)

        with pytest.raises(ValueError) as _:
            _, _ = po.tools.polar_to_rectangular(a, b)

    def test_coordinate_identity_transform_rectangular(self):
        dims = (10, 5, 256, 256)
        x = torch.randn(dims, device=DEVICE)
        y = torch.randn(dims, device=DEVICE)

        z = po.tools.polar_to_rectangular(
            *po.tools.rectangular_to_polar(torch.complex(x, y))
        )

        assert torch.linalg.vector_norm((x - z.real).flatten(), ord=2) < 1e-3
        assert torch.linalg.vector_norm((y - z.imag).flatten(), ord=2) < 1e-3

    def test_coordinate_identity_transform_polar(self):
        dims = (10, 5, 256, 256)

        # ensure vec len a is non-zero by adding .1 and then re-normalizing
        a = torch.rand(dims, device=DEVICE) + 0.1
        a = a / a.max()
        b = po.tools.rescale(torch.randn(dims, device=DEVICE), -pi / 2, pi / 2)

        A, B = po.tools.rectangular_to_polar(po.tools.polar_to_rectangular(a, b))

        assert torch.linalg.vector_norm((a - A).flatten(), ord=2) < 1e-3
        assert torch.linalg.vector_norm((b - B).flatten(), ord=2) < 1e-3

    @pytest.mark.parametrize(
        "size, expectation",
        [
            (500, pytest.raises(ValueError, match="output_size is bigger than")),
            (256, does_not_raise()),
            (128, does_not_raise()),
            (0, pytest.raises(ValueError, match="output_size must be positive")),
            (-10, pytest.raises(ValueError, match="output_size must be positive")),
            (10.0, pytest.raises(TypeError, match="output_size must be an int")),
            (torch.as_tensor(128), does_not_raise()),
            (np.asarray(128), does_not_raise()),
            (
                torch.as_tensor(128.0),
                pytest.raises(TypeError, match="output_size must be an int"),
            ),
            (
                [10, 10],
                pytest.raises(TypeError, match="output_size must be a single number"),
            ),
            (
                torch.as_tensor([10, 10]),
                pytest.raises(TypeError, match="output_size must be a single number"),
            ),
        ],
    )
    def test_center_crop(self, einstein_img, size, expectation):
        with expectation:
            po.tools.center_crop(einstein_img, size)

    @pytest.mark.parametrize("n", range(1, 15))
    def test_autocorrelation(self, n, basic_stim):
        x = basic_stim
        x_centered = x - x.mean((2, 3), keepdim=True)
        a = po.tools.autocorrelation(x_centered)
        a = po.tools.center_crop(a, n)

        # autocorr with zero delay is variance
        assert (
            torch.abs(torch.var(x, dim=(2, 3)) - a[..., n // 2, n // 2]) < 1e-5
        ).all()

        # autocorr can be computed in signal domain directly with roll
        h = randint(-(n // 2), ((n + 1) // 2))
        assert (
            torch.abs(
                (x_centered * torch.roll(x_centered, h, dims=2)).sum((2, 3))
                / (x.shape[-2] * x.shape[-1])
                - a[..., n // 2 + h, n // 2]
            )
            < 1e-5
        ).all()

        w = randint(-(n // 2), ((n + 1) // 2))
        assert (
            torch.abs(
                (x_centered * torch.roll(x_centered, w, dims=3)).sum((2, 3))
                / (x.shape[-2] * x.shape[-1])
                - a[..., n // 2, n // 2 + w]
            )
            < 1e-5
        ).all()

    @pytest.mark.parametrize("size_A", [1, 3])
    @pytest.mark.parametrize("size_B", [1, 2, 3])
    @pytest.mark.parametrize("dtype", [torch.float16, torch.float32, torch.float64])
    def test_add_noise(self, einstein_img, size_A, size_B, dtype):
        A = einstein_img.repeat(size_A, 1, 1, 1).to(dtype)
        B = size_B * [4]
        if size_A != size_B and size_A != 1 and size_B != 1:
            with pytest.raises(Exception):
                po.tools.add_noise(A, B)
        else:
            assert po.tools.add_noise(A, B).shape[0] == max(size_A, size_B)

    @pytest.mark.parametrize("factor", [0.5, 1, 1.5, 2, 1.1])
    @pytest.mark.parametrize("img_size", [256, 128, 200])
    def test_expand(self, factor, img_size, einstein_img):
        einstein_img = einstein_img.clone()[..., :img_size]
        if int(factor * img_size) != factor * img_size:
            expectation = pytest.raises(
                ValueError, match=r"factor \* x.shape\[-1\] must be"
            )
        elif int(factor * einstein_img.shape[-2]) != factor * einstein_img.shape[-2]:
            expectation = pytest.raises(
                ValueError, match=r"factor \* x.shape\[-2\] must be"
            )
        elif factor <= 1:
            expectation = pytest.raises(
                ValueError, match="factor must be strictly greater"
            )
        else:
            expectation = does_not_raise()
        with expectation:
            expanded = po.tools.expand(einstein_img, factor)
            np.testing.assert_equal(
                expanded.shape[-2:],
                [factor * s for s in einstein_img.shape[-2:]],
            )

    @pytest.mark.parametrize("factor", [0.5, 1, 1.5, 2, 1.1])
    @pytest.mark.parametrize("img_size", [256, 128, 200])
    def test_shrink(self, factor, img_size, einstein_img):
        einstein_img = einstein_img.clone()[..., :img_size]
        if int(img_size / factor) != img_size / factor:
            expectation = pytest.raises(
                ValueError, match=r"x.shape\[-1\]/factor must be"
            )
        elif int(einstein_img.shape[-2] / factor) != einstein_img.shape[-2] / factor:
            expectation = pytest.raises(
                ValueError, match=r"x.shape\[-2\]/factor must be"
            )
        elif factor <= 1:
            expectation = pytest.raises(
                ValueError, match="factor must be strictly greater"
            )
        else:
            expectation = does_not_raise()
        with expectation:
            shrunk = po.tools.shrink(einstein_img, factor)
            np.testing.assert_equal(
                shrunk.shape[-2:],
                [s / factor for s in einstein_img.shape[-2:]],
            )

    @pytest.mark.parametrize("batch_channel", [[1, 3], [2, 1], [2, 3]])
    def test_shrink_batch_channel(self, batch_channel, einstein_img):
        shrunk = po.tools.shrink(einstein_img.repeat((*batch_channel, 1, 1)), 2)
        size = batch_channel + [s / 2 for s in einstein_img.shape[-2:]]
        np.testing.assert_equal(shrunk.shape, size)

    @pytest.mark.parametrize("batch_channel", [[1, 3], [2, 1], [2, 3]])
    def test_expand_batch_channel(self, batch_channel, einstein_img):
        expanded = po.tools.expand(einstein_img.repeat((*batch_channel, 1, 1)), 2)
        size = batch_channel + [2 * s for s in einstein_img.shape[-2:]]
        np.testing.assert_equal(expanded.shape, size)

    @pytest.mark.parametrize("factor", [1.5, 2])
    @pytest.mark.parametrize("img", ["curie", "einstein", "metal", "nuts"])
    def test_expand_shrink(self, img, factor):
        # expand then shrink will be the same as the original image, up to this
        # fudge factor
        img = po.load_images(IMG_DIR / "256" / f"{img}.pgm").to(DEVICE)
        modified = po.tools.shrink(po.tools.expand(img, factor), factor)
        torch.testing.assert_close(img, modified, atol=2e-2, rtol=1e-6)

    @pytest.mark.parametrize("phase", [0, np.pi / 2, np.pi])
    def test_modulate_phase_correlation(self, phase):
        # here we create an image that has sinusoids at two frequencies, with
        # some phase offset. Because their frequencies are an octave apart,
        # they will show up in neighboring scales of the steerable pyramid
        # coefficients. Based on their phase offset, the coefficients should
        # either be correlated, uncorrelated, or anti-correlated, which is only
        # recoverable after using expand and doubling the phase of the lower
        # frequency one (this trick is used in th PS texture model)
        X = torch.arange(256).unsqueeze(1).repeat(1, 256) / 256 * 2 * torch.pi
        X = X.unsqueeze(0).unsqueeze(0)
        X = torch.sin(8 * X) + torch.sin(16 * X + phase)

        pyr = po.simul.SteerablePyramidFreq(X.shape[-2:], is_complex=True)
        pyr_coeffs = pyr(X)
        a = pyr_coeffs[(3, 2)]
        b = pyr_coeffs[(2, 2)]
        a = po.tools.expand(a, 2) / 4
        a = po.tools.modulate_phase(a, 2)

        # this is the correlation as computed in the PS texture model, which is
        # where modulate phase is used
        corr = einops.einsum(a.real, b.real, "b c h w, b c h w -> b c")
        corr = corr / (torch.mul(*a.shape[-2:])) / (a.std() * b.std())

        tgt_corr = {0: 0.4999, np.pi / 2: 0, np.pi: -0.4999}[phase]

        np.testing.assert_allclose(corr, tgt_corr, rtol=1e-5, atol=1e-5)

    def test_modulate_phase_noreal(self):
        X = torch.arange(256).unsqueeze(1).repeat(1, 256) / 256 * 2 * torch.pi
        X = X.unsqueeze(0).unsqueeze(0)

        with pytest.raises(TypeError, match="x must be a complex-valued tensor"):
            po.tools.modulate_phase(X)

    @pytest.mark.parametrize("batch_channel", [(1, 3), (2, 1), (2, 3)])
    def test_modulate_phase_batch_channel(self, batch_channel):
        X = torch.arange(256).unsqueeze(1).repeat(1, 256) / 256 * 2 * torch.pi
        X = X.unsqueeze(0).unsqueeze(0).repeat((*batch_channel, 1, 1))
        X = torch.sin(8 * X) + torch.sin(16 * X)

        pyr = po.simul.SteerablePyramidFreq(X.shape[-2:], is_complex=True)
        pyr_coeffs = pyr(X)
        a = pyr_coeffs[(3, 2)]
        a = po.tools.expand(a, 2) / 4
        a = po.tools.modulate_phase(a, 2)

        # shape should be preferred
        np.testing.assert_equal(a.shape[:2], batch_channel)

        # because the signal is just repeated along the batch and channel dims,
        # modulated version should be too (ensures we're not mixing batch or
        # channel)
        np.testing.assert_array_equal(a, a.roll(1, 1))
        np.testing.assert_array_equal(a, a.roll(1, 0))


class TestStats:
    def test_stats(self):
        torch.manual_seed(0)
        B, D = 32, 512
        x = torch.randn(B, D)
        m = torch.mean(x, dim=1, keepdim=True)
        v = po.tools.variance(x, mean=m, dim=1, keepdim=True)
        assert (
            torch.abs(v - torch.var(x, dim=1, keepdim=True, unbiased=False)) < 1e-5
        ).all()
        po.tools.skew(x, mean=m, var=v, dim=1)
        k = po.tools.kurtosis(x, mean=m, var=v, dim=1)
        assert torch.abs(k.mean() - 3) < 1e-1

        k = po.tools.kurtosis(torch.rand(B, D), dim=1)
        assert k.mean() < 3

        scale = 2
        exp_samples1 = -scale * torch.log(torch.rand(B, D))
        exp_samples2 = -scale * torch.log(torch.rand(B, D))
        lap_samples = exp_samples1 - exp_samples2
        k = po.tools.kurtosis(lap_samples, dim=1)
        assert k.mean() > 3

    @pytest.mark.parametrize("batch_channel", [(1, 1), (1, 3), (2, 1), (2, 3)])
    def test_var_multidim(self, batch_channel):
        B, D = 32, 512
        x = torch.randn(*batch_channel, B, D)
        var = po.tools.variance(x, dim=(-1, -2))
        np.testing.assert_equal(var.shape, batch_channel)

    @pytest.mark.parametrize("batch_channel", [(1, 1), (1, 3), (2, 1), (2, 3)])
    def test_skew_multidim(self, batch_channel):
        B, D = 32, 512
        x = torch.randn(*batch_channel, B, D)
        skew = po.tools.skew(x, dim=(-1, -2))
        np.testing.assert_equal(skew.shape, batch_channel)

    @pytest.mark.parametrize("batch_channel", [(1, 1), (1, 3), (2, 1), (2, 3)])
    def test_kurt_multidim(self, batch_channel):
        B, D = 32, 512
        x = torch.randn(*batch_channel, B, D)
        kurt = po.tools.kurtosis(x, dim=(-1, -2))
        np.testing.assert_equal(kurt.shape, batch_channel)


class TestDownsampleUpsample:
    @pytest.mark.parametrize("odd", [0, 1])
    @pytest.mark.parametrize("size", [9, 10, 11, 12])
    @pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
    @pytest.mark.parametrize("n_scales", [1, 2, 3])
    def test_filter(self, odd, size, dtype, n_scales):
        img = torch.zeros([1, 1, 48 + odd, 49], device=DEVICE, dtype=dtype)
        img[0, 0, 24, 24] = 1
        filt = np.zeros([size, size + 1])
        filt[5, 5] = 1
        filt = scipy.ndimage.gaussian_filter(filt, sigma=1)
        filt = torch.as_tensor(filt, dtype=dtype, device=DEVICE)
        img_down = po.tools.correlate_downsample(img, filt=filt)
        img_up = po.tools.upsample_convolve(img_down, odd=(odd, 1), filt=filt)
        assert np.unravel_index(img_up.cpu().numpy().argmax(), img_up.shape) == (
            0,
            0,
            24,
            24,
        )

        img_down = po.tools.blur_downsample(img, n_scales=n_scales)
        img_up = po.tools.upsample_blur(img_down, odd=(odd, 1), n_scales=n_scales)
        assert np.unravel_index(img_up.cpu().numpy().argmax(), img_up.shape) == (
            0,
            0,
            24,
            24,
        )

    @pytest.mark.parametrize("n_scales", [0, 1, 2, 3])
    @pytest.mark.parametrize("scale_filter", [True, False])
    def test_upsample(self, einstein_img, n_scales, scale_filter):
        if n_scales == 0:
            expectation = pytest.raises(ValueError, match="n_scales must be positive")
        else:
            expectation = does_not_raise()
        with expectation:
            us_img = po.tools.upsample_blur(
                einstein_img, (0, 0), n_scales, scale_filter=scale_filter
            )
            us_mn = us_img.mean()
            mn = einstein_img.mean()
            if us_img.shape[-2:] != (256 * 2**n_scales, 256 * 2**n_scales):
                raise Exception(f"upsampled shape is unexpected, {us_img.shape[-2:]}!")
            if scale_filter and not torch.isclose(us_mn, mn, atol=3e-3):
                raise Exception(
                    f"upsampled shape has unexpected mean, {us_mn} vs. {mn}!"
                )
            if not scale_filter and not torch.isclose(
                us_mn * (2**n_scales), mn, atol=1e-2
            ):
                raise Exception(
                    f"upsampled shape has unexpected mean, {us_mn} "
                    f"({us_mn * (2**n_scales)}) vs. {mn}!"
                )

    @pytest.mark.parametrize("n_scales", [0, 1, 2, 3])
    @pytest.mark.parametrize("scale_filter", [True, False])
    def test_downsample(self, einstein_img, n_scales, scale_filter):
        if n_scales == 0:
            expectation = pytest.raises(ValueError, match="n_scales must be positive")
        else:
            expectation = does_not_raise()
        with expectation:
            ds_img = po.tools.blur_downsample(
                einstein_img, n_scales, scale_filter=scale_filter
            )
            ds_mn = ds_img.mean()
            mn = einstein_img.mean()
            if ds_img.shape[-2:] != (256 // 2**n_scales, 256 // 2**n_scales):
                raise Exception(
                    f"downsampled shape is unexpected, {ds_img.shape[-2:]}!"
                )
            if scale_filter and not torch.isclose(ds_mn, mn, atol=3e-3):
                raise Exception(
                    f"downsampled shape has unexpected mean, {ds_mn} vs. {mn}!"
                )
            if not scale_filter and not torch.isclose(
                ds_mn, mn * (2**n_scales), atol=1e-2
            ):
                raise Exception(
                    f"downsampled shape has unexpected mean, {ds_mn} vs. {mn} "
                    f"({mn * (2**n_scales)})!"
                )

    def test_multichannel(self):
        img = torch.randn([10, 3, 24, 25], device=DEVICE, dtype=torch.float32)
        filt = torch.randn([5, 5], device=DEVICE, dtype=torch.float32)
        img_down = po.tools.correlate_downsample(img, filt=filt)
        img_up = po.tools.upsample_convolve(img_down, odd=(0, 1), filt=filt)
        assert img_up.shape == img.shape

        img_down = po.tools.blur_downsample(img)
        img_up = po.tools.upsample_blur(img_down, odd=(0, 1))
        assert img_up.shape == img.shape


class TestValidate:
    # https://docs.pytest.org/en/4.6.x/example/parametrize.html#parametrizing-conditional-raising
    @pytest.mark.parametrize(
        "shape,expectation",
        [
            ((1, 1, 16, 16), does_not_raise()),
            ((1, 3, 16, 16), does_not_raise()),
            ((2, 1, 16, 16), does_not_raise()),
            ((2, 3, 16, 16), does_not_raise()),
            (
                (1, 1, 1, 16, 16),
                pytest.warns(
                    UserWarning, match="methods have mostly been tested on 4d"
                ),
            ),
            (
                (1, 16, 16),
                pytest.warns(
                    UserWarning, match="methods have mostly been tested on 4d"
                ),
            ),
            (
                (16, 16),
                pytest.warns(
                    UserWarning, match="methods have mostly been tested on 4d"
                ),
            ),
        ],
    )
    def test_input_shape(self, shape, expectation):
        img = torch.rand(*shape)
        with expectation:
            po.tools.validate.validate_input(img)

    def test_input_no_batch(self):
        img = torch.rand(2, 1, 16, 16)
        with pytest.raises(ValueError, match="input_tensor batch dimension must be 1"):
            po.tools.validate.validate_input(img, no_batch=True)

    @pytest.mark.parametrize(
        "minmax,expectation",
        [
            (
                "min",
                pytest.raises(ValueError, match="input_tensor range must lie within"),
            ),
            (
                "max",
                pytest.raises(ValueError, match="input_tensor range must lie within"),
            ),
            (
                "range",
                pytest.raises(
                    ValueError,
                    match=r"allowed_range\[0\] must be strictly less",
                ),
            ),
        ],
    )
    def test_input_allowed_range(self, minmax, expectation):
        img = torch.rand(1, 1, 16, 16)
        allowed_range = (0, 1)
        if minmax == "min":
            img -= 1
        elif minmax == "max":
            img += 1
        elif minmax == "range":
            allowed_range = (1, 0)
        with expectation:
            po.tools.validate.validate_input(img, allowed_range=allowed_range)

    @pytest.mark.parametrize("model", ["frontend.OnOff"], indirect=True)
    def test_model_learnable(self, model):
        with pytest.raises(ValueError, match="model adds gradient to input"):
            po.tools.validate.validate_model(model, device=DEVICE)

    def test_model_numpy_comp(self):
        class TestModel(torch.nn.Module):
            def __init__(self):
                super().__init__()

            def forward(self, img):
                return np.fft.fft(img)

        model = TestModel()
        model.eval()
        with pytest.raises(
            ValueError, match="model does not return a torch.Tensor object"
        ):
            # don't pass device here because the model just uses numpy, which
            # only works on cpu
            po.tools.validate.validate_model(model)

    def test_model_detach(self):
        class TestModel(torch.nn.Module):
            def __init__(self):
                super().__init__()

            def forward(self, img):
                return img.detach()

        model = TestModel()
        model.eval()
        with pytest.raises(ValueError, match="model strips gradient from input"):
            po.tools.validate.validate_model(model, device=DEVICE)

    def test_model_numpy_and_back(self):
        class TestModel(torch.nn.Module):
            def __init__(self):
                super().__init__()

            def forward(self, img):
                return torch.from_numpy(np.fft.fft(img))

        model = TestModel()
        model.eval()
        with pytest.raises(
            ValueError,
            match="model tries to cast the input into something other",
        ):
            # don't pass device here because the model just uses numpy, which
            # only works on cpu
            po.tools.validate.validate_model(model)

    def test_model_precision(self):
        class TestModel(torch.nn.Module):
            def __init__(self):
                super().__init__()

            def forward(self, img):
                return img.to(torch.float16)

        model = TestModel()
        model.eval()
        with pytest.raises(TypeError, match="model changes precision of input"):
            po.tools.validate.validate_model(model, device=DEVICE)

    @pytest.mark.parametrize("model", ["diff_dims-2", "diff_dims-5"], indirect=True)
    def test_model_output_dim(self, model):
        model.eval()
        with pytest.warns(
            UserWarning, match="mostly been tested on models which produce 3d"
        ):
            po.tools.validate.validate_model(model, device=DEVICE)

    @pytest.mark.skipif(DEVICE.type == "cpu", reason="Only makes sense to test on cuda")
    def test_model_device(self):
        class TestModel(torch.nn.Module):
            def __init__(self):
                super().__init__()

            def forward(self, img):
                return img.to("cpu")

        model = TestModel()
        model.eval()
        with pytest.raises(RuntimeError, match="model changes device of input"):
            po.tools.validate.validate_model(model, device=DEVICE)

    @pytest.mark.parametrize("model", ["ColorModel"], indirect=True)
    def test_model_image_shape(self, model):
        img_shape = (1, 3, 16, 16)
        po.tools.validate.validate_model(model, image_shape=img_shape, device=DEVICE)

    @pytest.mark.filterwarnings(
        "ignore:Validating whether model can work with coarse-to-fine:UserWarning"
    )
    def test_validate_ctf_scales(self):
        class TestModel(torch.nn.Module):
            def __init__(self):
                super().__init__()

            def forward(self, img):
                return img

        model = TestModel()
        model.eval()
        with pytest.raises(AttributeError, match="model has no scales attribute"):
            po.tools.validate.validate_coarse_to_fine(model, device=DEVICE)

    @pytest.mark.filterwarnings(
        "ignore:Validating whether model can work with coarse-to-fine:UserWarning"
    )
    def test_validate_ctf_arg(self):
        class TestModel(torch.nn.Module):
            def __init__(self):
                super().__init__()
                self.scales = [0, 1, 2]

            def forward(self, img):
                return img

        model = TestModel()
        model.eval()
        with pytest.raises(
            TypeError,
            match="model forward method does not accept scales argument",
        ):
            po.tools.validate.validate_coarse_to_fine(model, device=DEVICE)

    @pytest.mark.filterwarnings(
        "ignore:Validating whether model can work with coarse-to-fine:UserWarning"
    )
    def test_validate_ctf_shape(self):
        class TestModel(torch.nn.Module):
            def __init__(self):
                super().__init__()
                self.scales = [0, 1, 2]

            def forward(self, img, scales=[]):
                return img

        model = TestModel()
        model.eval()
        with pytest.raises(
            ValueError,
            match="Output of model forward method doesn't change shape",
        ):
            po.tools.validate.validate_coarse_to_fine(model, device=DEVICE)

    @pytest.mark.parametrize(
        "model",
        ["PortillaSimoncelli"],
        indirect=True,
    )
    @pytest.mark.filterwarnings(
        "ignore:Validating whether model can work with coarse-to-fine:UserWarning"
    )
    def test_validate_ctf_pass(self, model):
        po.tools.validate.validate_coarse_to_fine(
            model, image_shape=(1, 1, *model.image_shape), device=DEVICE
        )

    # Metric validation tests
    def test_validate_metric_inputs(self):
        def identity_metric(x):
            return x

        with pytest.raises(TypeError, match="metric should be callable and accept two"):
            po.tools.validate.validate_metric(identity_metric, device=DEVICE)

    def test_validate_metric_output_shape(self):
        def difference_metric(x, y):
            return x - y

        with pytest.raises(
            ValueError, match="metric should return a scalar value but output"
        ):
            po.tools.validate.validate_metric(difference_metric, device=DEVICE)

    def test_validate_metric_identical(self):
        def mean_metric(x, y):
            return (x + y).mean()

        with pytest.raises(
            ValueError, match="metric should return <= 5e-7 on two identical"
        ):
            po.tools.validate.validate_metric(mean_metric, device=DEVICE)

    def test_validate_metric_nonnegative(self):
        po.tools.set_seed(0)

        def sum_metric(x, y):
            return (x - y).sum()

        with pytest.raises(
            ValueError, match="metric should always return non-negative"
        ):
            po.tools.validate.validate_metric(sum_metric, device=DEVICE)

    @pytest.mark.parametrize("model", ["frontend.OnOff.nograd"], indirect=True)
    def test_remove_grad(self, model):
        po.tools.validate.validate_model(model, device=DEVICE)


class TestOptim:
    def test_penalize_range_above(self):
        img = 0.5 * torch.ones((1, 1, 4, 4))
        img[..., 0, :] = 2
        assert po.tools.optim.penalize_range(img).item() == 4

    def test_penalize_range_below(self):
        img = 0.5 * torch.ones((1, 1, 4, 4))
        img[..., 0, :] = -1
        assert po.tools.optim.penalize_range(img).item() == 4


class TestPolarImages:
    def test_polar_angle_clockwise(self):
        ang = po.tools.polar_angle(100, direction="clockwise")
        idx = torch.argmin((ang - np.pi / 2) ** 2)
        assert torch.unravel_index(idx, (100, 100))[0] > 50, (
            "pi/2 should be in bottom half of image!"
        )
        idx = torch.argmin((ang + np.pi / 2) ** 2)
        assert torch.unravel_index(idx, (100, 100))[0] < 50, (
            "pi/2 should be in top half of image!"
        )

    def test_polar_angle_counterclockwise(self):
        ang = po.tools.polar_angle(100, direction="counter-clockwise")
        idx = torch.argmin((ang - np.pi / 2) ** 2)
        assert torch.unravel_index(idx, (100, 100))[0] < 50, (
            "pi/2 should be in top half of image!"
        )
        idx = torch.argmin((ang + np.pi / 2) ** 2)
        assert torch.unravel_index(idx, (100, 100))[0] > 50, (
            "pi/2 should be in bottom half of image!"
        )

    def test_polar_angle_direction(self):
        with pytest.raises(ValueError, match="direction must be one of"):
            po.tools.polar_angle(100, direction="-clockwise")


class TestEqualityChecks:
    def test_equal(self, einstein_img):
        _check_tensor_equality(einstein_img, einstein_img)

    def test_values(self, einstein_img, curie_img):
        with pytest.raises(ValueError, match="Different values"):
            _check_tensor_equality(einstein_img, curie_img)

    def test_dtype(self, einstein_img):
        with pytest.raises(ValueError, match="Different dtype"):
            _check_tensor_equality(einstein_img, einstein_img.to(torch.float64))

    def test_shape(self, einstein_img):
        with pytest.raises(ValueError, match="Different shape"):
            _check_tensor_equality(einstein_img, einstein_img[..., :64, :64])

    @pytest.mark.skipif(DEVICE.type == "cpu", reason="Only makes sense to test on cuda")
    def test_device(self, einstein_img):
        with pytest.raises(ValueError, match="Different device"):
            _check_tensor_equality(einstein_img.to("cuda"), einstein_img.to("cpu"))



"""Helper functions for testing."""

import pathlib
import re

import numpy as np
import pyrtools as pt
import torch

import plenoptic as po
from plenoptic.data.fetch import fetch_data
from test_models import TestPortillaSimoncelli


def update_ps_synthesis_test_file(
    torch_version: str | None = None,
) -> po.synthesize.metamer.Metamer:
    """Create new test file for test_models.test_ps_synthesis().

    We cannot guarantee perfect reproducibility across pytorch versions, but we
    can guarantee it within versions. Therefore, we need to periodically create
    new files to test our synthesis outputs against. This helper function does
    that, but note that you NEED to examine the resulting metamer manually to
    ensure it still looks good.

    After generating this file and checking it looks good, upload it to the OSF
    plenoptic-files project: https://osf.io/ts37w/files/osfstorage, then update
    test_models.get_portilla_simoncelli_synthesize_filename and
    fetch.REGISTRY_URLS and fetch.REGISTRY

    Parameters
    ----------
    torch_version
        The version of pytorch for which we should grab the corresponding test
        file. If ``None``, we use the installed version.

    Returns
    -------
    met
        Metamer object for inspection

    """
    ps_synth_file = fetch_data(
        "portilla_simoncelli_synthesize_torch_v1.12.0_ps-refactor-2.npz"
    )
    print(f"Loading from {ps_synth_file}")

    with np.load(ps_synth_file) as f:
        im = f["im"]
        im_init = f["im_init"]
        im_synth = f["im_synth"]
        rep_synth = f["rep_synth"]

    met = TestPortillaSimoncelli().test_ps_synthesis(ps_synth_file, False)

    torch_v = torch.__version__.split("+")[0]
    file_name_parts = re.findall(
        "(.*portilla_simoncelli_synthesize)(_gpu)?(_torch_v)?([0-9.]*)(_ps-refactor)?(-2)?.npz",
        ps_synth_file.name,
    )[0]
    output_file_name = (
        "".join(file_name_parts[:2]) + f"_torch_v{torch_v}{file_name_parts[-1]}.npz"
    )
    output = po.to_numpy(met.metamer).squeeze()
    rep = po.to_numpy(met.model(met.metamer)).squeeze()
    try:
        np.testing.assert_allclose(output, im_synth.squeeze(), rtol=1e-4, atol=1e-4)
        np.testing.assert_allclose(rep, rep_synth.squeeze(), rtol=1e-4, atol=1e-4)
        print(
            "Current synthesis same as saved version, so not saving current synthesis."
        )
    # only do all this if the tests would've failed
    except AssertionError:
        print(f"Saving at {output_file_name}")
        np.savez(
            output_file_name,
            im=im,
            im_init=im_init,
            im_synth=output,
            rep_synth=rep,
        )
        fig = pt.imshow(
            [output, im_synth.squeeze()],
            title=[f"New metamer (torch {torch_v})", "Old metamer"],
        )
        fig.savefig(output_file_name.replace(".npz", ".png"))
    return met


def update_ps_torch_output(save_dir):
    """Create new test files for test_ps_torch_output().

    Unlike update_ps_synthesis_test_file(), this only needs to get updated if we
    change the model.

    """
    save_dir = pathlib.Path(save_dir)
    save_dir.mkdir(parents=True, exist_ok=True)
    n_scales = [1, 2, 3, 4]
    n_orientations = [2, 3, 4]
    spatial_corr_width = range(3, 10)
    IMG_DIR = fetch_data("test_images.tar.gz")
    im_names = ["curie", "einstein", "metal", "nuts"]
    ims = po.load_images([IMG_DIR / "256" / f"{im}.pgm" for im in im_names])
    for scale in n_scales:
        for ori in n_orientations:
            for width in spatial_corr_width:
                for im, name in zip(ims, im_names):
                    ps = po.simul.PortillaSimoncelli(
                        im.shape[-2:],
                        n_scales=scale,
                        n_orientations=ori,
                        spatial_corr_width=width,
                    ).to(torch.float64)
                    output = po.to_numpy(ps(im.unsqueeze(0)))
                    fname = (
                        save_dir / f"{name}_scales-{scale}_ori-{ori}_spat-{width}.npy"
                    )
                    np.save(fname, output)
    print(
        f"All outputs have been saved in directory {save_dir},"
        + f"now go to {save_dir.parent} "
        f"and run `tar czf {save_dir.name}"
        + f" --directory={save_dir.with_suffix('.tar.gz').name}/ .`"
    )


def update_ps_scales(save_path):
    """Create new test files for test_ps_torch_output().

    Unlike update_ps_synthesis_test_file(), this only needs to get updated if we
    change the model.

    """
    save_path = pathlib.Path(save_path)
    if save_path.suffix != ".npz":
        raise ValueError(f"save_path must have suffix .npz but got {save_path.suffix}!")
    save_path.parent.mkdir(parents=True, exist_ok=True)
    n_scales = [1, 2, 3, 4]
    n_orientations = [2, 3, 4]
    spatial_corr_width = range(3, 10)
    output = {}
    for scale in n_scales:
        for ori in n_orientations:
            for width in spatial_corr_width:
                ps = po.simul.PortillaSimoncelli(
                    (256, 256),
                    n_scales=scale,
                    n_orientations=ori,
                    spatial_corr_width=width,
                )
                key = f"scale-{scale}_ori-{ori}_width-{width}"
                output[key] = ps._representation_scales
    np.savez(save_path, **output)



"""
Eigendistortions.

Classes to perform the synthesis of eigendistortions.
"""

import warnings
from collections.abc import Callable
from typing import Any, Literal

import matplotlib.pyplot
import torch
from matplotlib.figure import Figure
from torch import Tensor
from tqdm.auto import tqdm

from ..tools.display import imshow
from ..tools.validate import validate_input, validate_model
from .autodiff import (
    jacobian,
    jacobian_vector_product,
    vector_jacobian_product,
)
from .synthesis import Synthesis


def fisher_info_matrix_vector_product(
    y: Tensor, x: Tensor, v: Tensor, dummy_vec: Tensor
) -> Tensor:
    r"""
    Compute Fisher Information Matrix Vector Product: :math:`Fv`.

    Parameters
    ----------
    y
        Output tensor with gradient attached.
    x
        Input tensor with gradient attached.
    v
        The vectors with which to compute Fisher vector products.
    dummy_vec
        Dummy vector for Jacobian vector product trick.

    Returns
    -------
    Fv
        Vector, Fisher vector product.

    Notes
    -----
    Under white Gaussian noise assumption, :math:`F` is matrix multiplication
    of Jacobian transpose and Jacobian:
    :math:`F = J^T J`. Hence:
    :math:`Fv = J^T (Jv)`
    """  # numpydoc ignore=ES01
    Jv = jacobian_vector_product(y, x, v, dummy_vec)
    Fv = vector_jacobian_product(y, x, Jv, detach=True)

    return Fv


def fisher_info_matrix_eigenvalue(
    y: Tensor, x: Tensor, v: Tensor, dummy_vec: Tensor | None = None
) -> Tensor:
    r"""
    Compute eigenvalues of Fisher vector products.

    We compute the Fisher Information Matrix corresponding to eigenvectors in ``v``:
    :math:`\lambda= v^T F v`.

    Parameters
    ----------
    y
        Output tensor with gradient attached.
    x
        Input tensor with gradient attached.
    v
        The vectors with which to compute Fisher vector products.
    dummy_vec
        Dummy vector for Jacobian vector product trick.

    Returns
    -------
    lmbda
        The computed eigenvalues.
    """
    if dummy_vec is None:
        dummy_vec = torch.ones_like(y, requires_grad=True)

    Fv = fisher_info_matrix_vector_product(y, x, v, dummy_vec)

    # compute eigenvalues for all vectors in v
    lmbda = torch.stack([a.dot(b) for a, b in zip(v.T, Fv.T)])
    return lmbda


class Eigendistortion(Synthesis):
    r"""
    Synthesize eigendistortions induced by a model on a given input image.

    Following the basic idea in [1]_, this class synthesizes image perturbations that
    are considered the most and least noticeable for a model on a given image. Because
    these are perturbations on the input image, they are local in pixel space, i.e.,
    they do not change the pixels much.

    Parameters
    ----------
    image
        Image to perturb. We currently do not support batches of images, as each image
        requires its own optimization, so either ``image.ndimension()==1`` or
        ``image.shape[0]==1``.
    model
        Torch model with defined forward and backward operations.

    Notes
    -----
    This is a method for comparing image representations in terms of their ability to
    explain perceptual sensitivity in humans. It estimates eigenvectors of the Fisher
    Information Matrix. A model, :math:`y = f(x)`, is a deterministic (and
    differentiable) mapping from the input pixels :math:`x \in \mathbb{R}^n` to a mean
    output response vector :math:`y\in \mathbb{R}^m`, where we assume additive white
    Gaussian noise in the response space.
    The Jacobian matrix at :math:`x` is:
        :math:`J(x) = J = dydx`,
        :math:`J\in\mathbb{R}^{m \times n}` (ie. output_dim x input_dim)
    The matrix consists of all partial derivatives of the vector-valued function
    :math:`f`. The Fisher Information Matrix (FIM) at :math:`x`, under white Gaussian
    noise in the response space, is:
        :math:`F = J^T J`
    It is a quadratic approximation of the discriminability of distortions
    relative to :math:`x`.

    References
    ----------
    .. [1] Berardino, A., Laparra, V., Ballé, J. and Simoncelli, E., 2017.
           Eigen-distortions of hierarchical representations. In Advances in
           neural information processing systems (pp. 3530-3539).
           https://www.cns.nyu.edu/pub/lcv/berardino17c-final.pdf
           https://www.cns.nyu.edu/~lcv/eigendistortions/
    """

    def __init__(self, image: Tensor, model: torch.nn.Module):
        super().__init__()
        validate_input(image, no_batch=True)
        validate_model(
            model,
            image_shape=image.shape,
            image_dtype=image.dtype,
            device=image.device,
        )

        self._image_shape = image.shape
        if image.ndimension() != 1:
            self._image_shape = self._image_shape[1:]

        self._model = model
        # flatten and attach gradient and reshape to image
        self._image_flat = image.flatten().unsqueeze(1).requires_grad_(True)
        self._init_representation(image)

        self._jacobian = None
        self._eigendistortions = None
        self._eigenvalues = None
        self._eigenindex = None

    def _init_representation(self, image: torch.Tensor):
        """Set self._representation_flat, based on ``self.model`` and ``image``."""
        # numpydoc ignore=ES01,PR01
        self._image = self._image_flat.view(*image.shape)
        image_representation = self.model(self.image)

        if len(image_representation) > 1:
            self._representation_flat = torch.cat(
                [s.squeeze().view(-1) for s in image_representation]
            ).unsqueeze(1)
        else:
            self._representation_flat = (
                image_representation.squeeze().view(-1).unsqueeze(1)
            )

    def synthesize(
        self,
        method: Literal["exact", "power", "randomized_svd"] = "power",
        k: int = 1,
        max_iter: int = 1000,
        p: int = 5,
        q: int = 2,
        stop_criterion: float = 1e-7,
    ):
        r"""
        Compute eigendistortions of Fisher Information Matrix with given input image.

        There are three potential ways of computing the eigendistortion for a model;
        all have the same interpretation. See ``method`` argument for details.

        Parameters
        ----------
        method
            Eigensolver method. ``'exact'`` tries to do eigendecomposition directly
            (not recommended for very large inputs). ``'power'`` uses the power
            method to compute first and last eigendistortions, with maximum number of
            iterations dictated by n_steps. ``'randomized_svd'`` uses randomized SVD to
            approximate the top k eigendistortions and their corresponding eigenvalues.
        k
            How many vectors to return using block power method or svd.
        max_iter
            Maximum number of steps to run for ``method='power'`` in eigenvalue
            computation. Ignored for other methods.
        p
            Oversampling parameter for randomized SVD. k+p vectors will be sampled,
            and k will be returned. See docstring of :meth:`_synthesize_randomized_svd`
            for more details including algorithm reference.
        q
            Matrix power parameter for randomized SVD. This is an effective trick for
            the algorithm to converge to the correct eigenvectors when the
            eigenspectrum does not decay quickly. See :meth:`_synthesize_randomized_svd`
            for more details including algorithm reference.
        stop_criterion
            Used if ``method='power'`` to check for convergence. If the L2-norm
            of the eigenvalues has changed by less than this value from one
            iteration to the next, we terminate synthesis.

        Raises
        ------
        ValueError
            If ``method`` takes an illegal value.

        Warns
        -----
        UserWarning
            If ``method == "power"`` but the Jacobian size is greater than 1e6 (which
            depends on the number of elements in the model's representation and input
            image), in which case we're worried about running out of memory.
        """
        allowed_methods = ["power", "exact", "randomized_svd"]
        if method not in allowed_methods:
            raise ValueError(f"method must be in {allowed_methods}")

        if (
            method == "exact"
            and self._representation_flat.size(0) * self._image_flat.size(0) > 1e6
        ):
            warnings.warn(
                "Jacobian > 1e6 elements and may cause out-of-memory. Use"
                " method =  {'power', 'randomized_svd'}."
            )

        if method == "exact":  # compute exact Jacobian
            eig_vals, eig_vecs = self._synthesize_exact()
            eig_vecs = self._vector_to_image(eig_vecs.detach())
            eig_vecs_ind = torch.arange(len(eig_vecs))

        elif method == "randomized_svd":
            lmbda_new, v_new, error_approx = self._synthesize_randomized_svd(
                k=k, p=p, q=q
            )
            eig_vecs = self._vector_to_image(v_new.detach())
            eig_vals = lmbda_new.squeeze()
            eig_vecs_ind = torch.arange(k)

            # display the approximate estimation error of the range space
            warnings.warn(
                "Randomized SVD complete! Estimated spectral approximation"
                f" error = {error_approx:.2f}"
            )

        else:  # method == 'power'
            assert max_iter > 0, "max_iter must be greater than zero"

            lmbda_max, v_max = self._synthesize_power(
                k=k, shift=0.0, tol=stop_criterion, max_iter=max_iter
            )
            lmbda_min, v_min = self._synthesize_power(
                k=k, shift=lmbda_max[0], tol=stop_criterion, max_iter=max_iter
            )
            n = v_max.shape[0]

            eig_vecs = self._vector_to_image(torch.cat((v_max, v_min), dim=1).detach())
            eig_vals = torch.cat([lmbda_max, lmbda_min]).squeeze()
            eig_vecs_ind = torch.cat((torch.arange(k), torch.arange(n - k, n)))

        # reshape to (n x num_chans x h x w)
        self._eigendistortions = torch.stack(eig_vecs, 0) if len(eig_vecs) != 0 else []
        self._eigenvalues = torch.abs(eig_vals.detach())
        self._eigenindex = eig_vecs_ind

    def _synthesize_exact(self) -> tuple[Tensor, Tensor]:
        r"""
        Eigendecomposition of explicitly computed Fisher Information Matrix.

        To be used when the input is small (e.g. less than 70x70 image on cluster or
        30x30 on your own machine). This method obviates the power iteration and its
        related algorithms (e.g. Lanczos). This method computes the Fisher Information
        Matrix by explicitly computing the Jacobian of the representation with respect
        to the input.

        Returns
        -------
        eig_vals
            Eigenvalues in decreasing order.
        eig_vecs
            Eigenvectors in 2D tensor, whose cols are eigenvectors (i.e.
            eigendistortions) corresponding to eigenvalues.
        """
        J = self.compute_jacobian()
        F = J.T @ J
        eig_vals, eig_vecs = torch.linalg.eigh(F, UPLO="U")
        eig_vecs = eig_vecs.flip(dims=(1,))
        eig_vals = eig_vals.flip(dims=(0,))
        return eig_vals, eig_vecs

    def compute_jacobian(self) -> Tensor:
        r"""
        Compute (via :func:`autodiff.jacobian`), cache, and return jacobian.

        Returns
        -------
        J
            Jacobian of representation with respect to input.

        Warns
        -----
        UserWarning
            If input dimensionality is greater than 1e4, in which case we believe that
            this calculation will take too long.
        """  # numpydoc ignore=ES01
        if self.jacobian is None:
            J = jacobian(self._representation_flat, self._image_flat)
            self._jacobian = J
        else:
            J = self.jacobian

        return J

    def _synthesize_power(
        self, k: int, shift: Tensor | float, tol: float, max_iter: int
    ) -> tuple[Tensor, Tensor]:
        r"""
        Use power method  to obtain largest (smallest) eigenvalue/vector pairs.

        When ``k>1``, uses orthogonal iteration, see [1]_.

        Apply the algorithm to approximate the extremal eigenvalues and eigenvectors
        of the Fisher Information Matrix, without explicitly representing that matrix.

        This method repeatedly calls :func:`fisher_info_matrix_vector_product` with a
        single (``k=1``), or multiple (``k>1``) vectors.

        Parameters
        ----------
        k
            Number of top and bottom eigendistortions to synthesize; i.e. if k=2,
            then the top 2 and bottom 2 will be returned. When `k>1`, multiple
            eigendistortions are synthesized, and each power iteration step is followed
            by a QR orthogonalization step to ensure the vectors are orthonormal.
        shift
            When `shift=0`, this function estimates the top `k` eigenvalue/vector
            pairs. When `shift` is set to the estimated top eigenvalue this function
            will estimate the smallest eigenval/eigenvector pairs.
        tol
            Tolerance value.
        max_iter
            Maximum number of steps.

        Returns
        -------
        lmbda
            Eigenvalue corresponding to final vector of power iteration.
        v
            Final eigenvector(s) (i.e. eigendistortions) of power (orthogonal)
            iteration procedure.

        References
        ----------
        .. [1] Orthogonal iteration; Algorithm 8.2.8 Golub and Van Loan, Matrix
               Computations, 3rd Ed.
        """
        x, y = self._image_flat, self._representation_flat

        # note: v is an n x k matrix where k is number of eigendists to be synthesized!
        v = torch.randn(len(x), k, device=x.device, dtype=x.dtype)
        v = v / torch.linalg.vector_norm(v, dim=0, keepdim=True, ord=2)

        _dummy_vec = torch.ones_like(y, requires_grad=True)  # cache a dummy vec for jvp
        Fv = fisher_info_matrix_vector_product(y, x, v, _dummy_vec)
        v = Fv / torch.linalg.vector_norm(Fv, dim=0, keepdim=True, ord=2)
        lmbda = fisher_info_matrix_eigenvalue(y, x, v, _dummy_vec)

        d_lambda = torch.as_tensor(float("inf"))
        lmbda_new, v_new = None, None
        desc = ("Top" if shift == 0 else "Bottom") + f" k={k} eigendists"
        pbar = tqdm(range(max_iter), desc=desc)
        postfix_dict = {"delta_eigenval": None}

        for _ in pbar:
            postfix_dict.update(dict(delta_eigenval=f"{d_lambda.item():.2E}"))
            pbar.set_postfix(**postfix_dict)

            if d_lambda <= tol:
                print(f"{desc} computed | Stop criterion {tol:.2E} reached.")
                break

            Fv = fisher_info_matrix_vector_product(y, x, v, _dummy_vec)
            Fv = Fv - shift * v  # optionally shift: (F - shift*I)v

            v_new, _ = torch.linalg.qr(Fv, "reduced")  # (ortho)normalize vector(s)

            lmbda_new = fisher_info_matrix_eigenvalue(y, x, v_new, _dummy_vec)

            d_lambda = torch.linalg.vector_norm(
                lmbda - lmbda_new, ord=2
            )  # stability of eigenspace
            v = v_new
            lmbda = lmbda_new

        pbar.close()

        return lmbda_new, v_new

    def _synthesize_randomized_svd(
        self, k: int, p: int, q: int
    ) -> tuple[Tensor, Tensor, Tensor]:
        r"""
        Synthesize eigendistortions using randomized truncated SVD.

        This method approximates the column space of the Fisher Info Matrix, projects
        the FIM into that column space, then computes its SVD. See [1]_ for details.

        Parameters
        ----------
        k
            Number of eigenvecs (rank of factorization) to be returned.
        p
            Oversampling parameter, recommended to be 5.
        q
            Matrix power iteration. Used to squeeze the eigen spectrum for more
            accurate approximation. Recommended to be 2.

        Returns
        -------
        S
            Eigenvalues, Size((n, )).
        V
            Eigendistortions, Size((n, k)).
        error_approx
            Estimate of the approximation error. Defined as the expected error
            between the true subspace and approximated subspace.

        References
        ----------
        .. [1] Halko, Martinsson, Tropp, Finding structure with randomness:
               Probabilistic algorithms for constructing approximate matrix
               decompositions, SIAM Rev. 53:2, pp. 217-288
               https://arxiv.org/abs/0909.4061 (2011)
        """
        x, y = self._image_flat, self._representation_flat
        n = len(x)

        P = torch.randn(n, k + p).to(x.device)
        # orthogonalize first for numerical stability
        P, _ = torch.linalg.qr(P, "reduced")
        _dummy_vec = torch.ones_like(y, requires_grad=True)
        Z = fisher_info_matrix_vector_product(y, x, P, _dummy_vec)

        # optional power iteration to squeeze the spectrum for more accurate
        # estimate
        for _ in range(q):
            Z = fisher_info_matrix_vector_product(y, x, Z, _dummy_vec)

        Q, _ = torch.linalg.qr(Z, "reduced")
        # B = Q.T @ A @ Q
        B = Q.T @ fisher_info_matrix_vector_product(y, x, Q, _dummy_vec)
        _, S, Vh = torch.linalg.svd(B, False)  # eigendecomp of small matrix
        V = Vh.T
        V = Q @ V  # lift up to original dimensionality

        # estimate error in Q estimate of range space
        omega = fisher_info_matrix_vector_product(
            y, x, torch.randn(n, 20).to(x.device), _dummy_vec
        )
        error_approx = omega - (Q @ Q.T @ omega)
        error_approx = torch.linalg.vector_norm(error_approx, dim=0, ord=2).mean()

        return S[:k].clone(), V[:, :k].clone(), error_approx  # truncate

    def _vector_to_image(self, vecs: Tensor) -> list[Tensor]:
        r"""
        Reshape eigenvectors back into correct image dimensions.

        Parameters
        ----------
        vecs
            Eigendistortion tensor with ``torch.Size([N, num_distortions])``.
            Each distortion will be reshaped into the original image shape and
            placed in a list.

        Returns
        -------
        imgs
            List of Tensor images.
        """  # numpydoc ignore=ES01
        imgs = [vecs[:, i].reshape(self._image_shape) for i in range(vecs.shape[1])]
        return imgs

    def _indexer(self, idx: int) -> int:
        """
        Map eigenindex to arg index (0-indexed).

        Parameters
        ----------
        idx
            Eigenvalue number.

        Returns
        -------
        idx
            Index into eigenvalue tensor.

        Raises
        ------
        ValueError
            If ``eigenindex`` doesn't correspond to one of the synthesized
            eigendistortions.
        ValueError
            If no eigendistortions have been synthesized.
        """  # numpydoc ignore=ES01
        n = len(self._image_flat)
        idx_range = range(n)
        i = idx_range[idx]

        all_idx = self.eigenindex
        if i not in all_idx:
            raise ValueError("eigenindex must be the index of one of the vectors")
        if all_idx is None or len(all_idx) == 0:
            raise ValueError("No eigendistortions synthesized")
        return torch.where(all_idx == i)[0].item()

    def save(self, file_path: str):
        r"""
        Save all relevant variables in .pt file.

        See :meth:`load` docstring for an example of use.

        Parameters
        ----------
        file_path : str
            The path to save the Eigendistortion object to.
        """
        save_io_attrs = [("_model", ("_image",))]
        super().save(file_path, save_io_attrs)

    def to(self, *args: Any, **kwargs: Any):
        r"""
        Move and/or cast the parameters and buffers.

        This can be called as

        .. function:: to(device=None, dtype=None, non_blocking=False)

        .. function:: to(dtype, non_blocking=False)

        .. function:: to(tensor, non_blocking=False)

        Its signature is similar to :meth:`torch.Tensor.to`, but only accepts
        floating point desired :attr:`dtype` s. In addition, this method will
        only cast the floating point parameters and buffers to :attr:`dtype`
        (if given). The integral parameters and buffers will be moved
        :attr:`device`, if that is given, but with dtypes unchanged. When
        :attr:`non_blocking` is set, it tries to convert/move asynchronously
        with respect to the host if possible, e.g., moving CPU Tensors with
        pinned memory to CUDA devices.

        See :meth:`torch.nn.Module.to` for examples.

        .. note::
            This method modifies the module in-place.

        Parameters
        ----------
        device : torch.device
            The desired device of the parameters and buffers in this module.
        dtype : torch.dtype
            The desired floating point type of the floating point parameters and
            buffers in this module.
        tensor : torch.Tensor
            Tensor whose dtype and device are the desired dtype and device for
            all parameters and buffers in this module.
        """  # numpydoc ignore=PR01,PR02
        attrs = [
            "_jacobian",
            "_eigendistortions",
            "_eigenvalues",
            "_eigenindex",
            "_image",
            "_image_flat",
            "_representation_flat",
        ]
        super().to(*args, attrs=attrs, **kwargs)
        # try to call .to() on model. this should work, but it might fail if e.g., this
        # a custom model that doesn't inherit torch.nn.Module
        try:
            self._model = self._model.to(*args, **kwargs)
        except AttributeError:
            warnings.warn("Unable to call model.to(), so we leave it as is.")
        # we need _representation_flat and _image_flat to be connected in the
        # computation graph for the autograd calls to work, so we reinitialize
        # it here
        self._init_representation(self.image)

    def load(
        self,
        file_path: str,
        map_location: str | None = None,
        tensor_equality_atol: float = 1e-8,
        tensor_equality_rtol: float = 1e-5,
        **pickle_load_args: Any,
    ):
        r"""
        Load all relevant stuff from a .pt file.

        This must be called by a ``Eigendistortion`` object initialized just like the
        saved object.

        Note this operates in place and so doesn't return anything.

        .. versionchanged:: 1.2
           load behavior changed in a backwards-incompatible manner in order to
           compatible with breaking changes in torch 2.6.

        Parameters
        ----------
        file_path
            The path to load the synthesis object from.
        map_location
            Argument to pass to :func:`torch.load` as ``map_location``. If you
            save stuff that was being run on a GPU and are loading onto a
            CPU, you'll need this to make sure everything lines up
            properly. This should be structured like the str you would
            pass to :class:`torch.device`.
        tensor_equality_atol
            Absolute tolerance to use when checking for tensor equality during load,
            passed to :func:`torch.allclose`. It may be necessary to increase if you are
            saving and loading on two machines with torch built by different cuda
            versions. Be careful when changing this! See
            :class:`torch.finfo<torch.torch.finfo>` for more details about floating
            point precision of different data types (especially, ``eps``); if you have
            to increase this by more than 1 or 2 decades, then you are probably not
            dealing with a numerical issue.
        tensor_equality_rtol
            Relative tolerance to use when checking for tensor equality during load,
            passed to :func:`torch.allclose`. It may be necessary to increase if you are
            saving and loading on two machines with torch built by different cuda
            versions. Be careful when changing this! See
            :class:`torch.finfo<torch.torch.finfo>` for more details about floating
            point precision of different data types (especially, ``eps``); if you have
            to increase this by more than 1 or 2 decades, then you are probably not
            dealing with a numerical issue.
        **pickle_load_args
            Any additional kwargs will be added to ``pickle_module.load`` via
            :func:`torch.load`, see that function's docstring for details.

        Raises
        ------
        ValueError
            If :func:`synthesize` has been called before this call to ``load``.
        ValueError
            If the object saved at ``file_path`` is not a ``Eigendistortion`` object.
        ValueError
            If the saved and loading ``Eigendistortion`` objects have a different value
            for :attr:`image`.
        ValueError
            If the behavior of :attr:`model` is different between the saved and loading
            objects.

        See Also
        --------
        :func:`~plenoptic.tools.io.examine_saved_synthesis`
            Examine metadata from saved object: pytorch and plenoptic versions, name of
            the synthesis object, shapes of tensors, etc.

        Examples
        --------
        >>> import plenoptic as po
        >>> img = po.data.einstein()
        >>> model = po.simul.Gaussian(30).eval()
        >>> po.tools.remove_grad(model)
        >>> eig = po.synth.Eigendistortion(img, model)
        >>> eig.synthesize(max_iter=5)
        >>> eig.save("eig.pt")
        >>> eig_copy = po.synth.Eigendistortion(img, model)
        >>> eig_copy.load("eig.pt")
        """
        check_attributes = ["_image"]
        check_io_attrs = [("_model", ("_image",))]
        super().load(
            file_path,
            "eigenindex",
            map_location=map_location,
            check_attributes=check_attributes,
            check_io_attributes=check_io_attrs,
            tensor_equality_atol=tensor_equality_atol,
            tensor_equality_rtol=tensor_equality_rtol,
            **pickle_load_args,
        )
        # make these require a grad again
        self._image_flat.requires_grad_()
        # we need _representation_flat and _image_flat to be connected in the
        # computation graph for the autograd calls to work, so we reinitialize
        # it here
        self._init_representation(self.image)

    @property
    def model(self) -> torch.nn.Module:
        """The model for which the eigendistortions are synthesized."""
        # numpydoc ignore=RT01,ES01
        return self._model

    @property
    def image(self) -> torch.Tensor:
        """Target image of eigendistortion synthesis."""
        # numpydoc ignore=RT01,ES01
        return self._image

    @property
    def jacobian(self) -> torch.Tensor:
        """
        Jacobian matrix of :attr:`model` with respect to :attr:`image`.

        Only set when :func:`synthesize` is run with ``method='exact'``.
        Else, ``None``.
        """  # numpydoc ignore=RT01
        return self._jacobian

    @property
    def eigendistortions(self) -> torch.Tensor:
        """
        Eigendistortions, ordered by eigenvalue.

        Eigendistortions are the eigenvectors of Fisher matrix, will have size
        ``Size((n_distortions, *image.shape[1:]))``.
        """  # numpydoc ignore=RT01
        return self._eigendistortions

    @property
    def eigenvalues(self) -> torch.Tensor:
        """Eigenvalues corresponding to each eigendistortion, in decreasing order."""
        # numpydoc ignore=RT01,ES01,SS05
        return self._eigenvalues

    @property
    def eigenindex(self) -> torch.Tensor:
        """Index of each eigenvector/eigenvalue."""
        # numpydoc ignore=RT01,ES01
        return self._eigenindex


def display_eigendistortion(
    eigendistortion: Eigendistortion,
    eigenindex: int = 0,
    alpha: float = 5.0,
    process_image: Callable[[Tensor], Tensor] = lambda x: x,
    ax: matplotlib.axes.Axes | None = None,
    plot_complex: str = "rectangular",
    **kwargs: Any,
) -> Figure:
    r"""
    Display specified eigendistortion added to the image.

    If image or eigendistortions have 3 channels, then it is assumed to be a color
    image and it is converted to grayscale. This is merely for display convenience
    and may change in the future.

    Parameters
    ----------
    eigendistortion
        Eigendistortion object whose synthesized eigendistortion we want to display.
    eigenindex
        Index of eigendistortion to plot. E.g. If there are 10 eigenvectors, 0 will
        index the first one, and -1 or 9 will index the last one.
    alpha
        Amount by which to scale eigendistortion for
        ``image + (alpha * eigendistortion)`` for display.
    process_image
        A function to process the image+alpha*distortion before clamping between 0,1.
        E.g. multiplying by the stdev ImageNet then adding the mean of ImageNet to undo
        image preprocessing.
    ax
        Axis handle on which to plot.
    plot_complex
        Parameter for :func:`~plenoptic.tools.display.imshow` determining how to handle
        complex values. See that method's docstring for details.
    **kwargs
        Additional arguments for :func:`~plenoptic.tools.display.imshow`.

    Returns
    -------
    fig
        Figure containing the displayed images.

    Raises
    ------
    ValueError
        If ``eigenindex`` doesn't correspond to one of the synthesized eigendistortions.

    See Also
    --------
    display_eigendistortion_all
        Display base image and multiple eigendistortions, alone and added to image.
    """
    # reshape so channel dim is last
    im_shape = eigendistortion._image_shape
    image = eigendistortion.image.detach().view(1, *im_shape).cpu()
    dist = (
        eigendistortion.eigendistortions[eigendistortion._indexer(eigenindex)]
        .unsqueeze(0)
        .cpu()
    )

    img_processed = process_image(image + alpha * dist)
    to_plot = torch.clamp(img_processed, 0, 1)
    title = f"{alpha} * Eigendistortion[{eigenindex}]"
    fig = imshow(to_plot, ax=ax, plot_complex=plot_complex, title=title, **kwargs)

    return fig


def display_eigendistortion_all(
    eigendistortion: Eigendistortion,
    eigenindex: int | list[int] = [0, -1],
    alpha: float | list[float] = 5.0,
    process_image: Callable[[Tensor], Tensor] | None = None,
    plot_complex: str = "rectangular",
    as_rgb: bool = False,
    suptitle: str = "Eigendistortions",
    suptitle_kwargs: dict | None = None,
    **kwargs: Any,
) -> Figure:
    r"""
    Display base image, eigendistortions alone, and eigendistortions added to the image.

    If image or eigendistortions have 3 channels, then it is assumed to be a color
    image and it is converted to grayscale. This is merely for display convenience
    and may change in the future.

    Parameters
    ----------
    eigendistortion
        Eigendistortion object whose synthesized eigendistortion we want to display.
    eigenindex
        Index of eigendistortion to plot. E.g. If there are 10 eigenvectors, 0 will
        index the first one, and -1 or 9 will index the last one.
    alpha
        Amount by which to scale eigendistortion for ``image + (alpha *
        eigendistortion)`` for display. If a list, must be the same length as
        ``eigenindex`` and will multiply each distortion by the corresponding
        ``alpha`` value.
    process_image
        A function to process all images before display. E.g. multiplying by the
        stdev ImageNet then adding the mean of ImageNet to undo image
        preprocessing. If ``None`` and ``as_rgb is True``, will add 0.5 to the
        distortion(s) (to avoid matplotlib clipping), else if ``None`` do nothing.
    plot_complex
        Parameter for :func:`~plenoptic.tools.display.imshow` determining how to handle
        complex values. See that method's docstring for details.
    as_rgb
        Whether to consider the channels as encoding RGB(A) values. If ``True``, we
        attempt to plot the image in color, so your tensor must have 3 (or 4 if
        you want the alpha channel) elements in the channel dimension. If ``False``,
        we plot each channel as a separate grayscale image.
    suptitle
        Super title to plot above all axes.
    suptitle_kwargs
        Additional arguments for :func:`matplotlib.pyplot.suptitle`.
    **kwargs
        Additional arguments for :func:`~plenoptic.tools.display.imshow`.

    Returns
    -------
    fig
        Figure containing the displayed images.

    Raises
    ------
    ValueError
        If ``len(alpha) != len(eigenindex)``.
    ValueError
        If a value of ``eigenindex`` doesn't correspond to one of the
        synthesized eigendistortions.

    Warns
    -----
    UserWarning
        If ``process_image=None`` and ``as_rgb=True``, because we are adding 0.5
        to the distortion.

    See Also
    --------
    display_eigendistortion
        Display single eigendistortion added to image.

    Examples
    --------
    .. plot::

      >>> import plenoptic as po
      >>> import torch
      >>> from plenoptic.data.fetch import fetch_data
      >>> img = po.data.einstein().to(torch.float64)
      >>> lg = po.simul.LuminanceGainControl(
      ...     (31, 31), pad_mode="circular", pretrained=True, cache_filt=True
      ... ).eval()
      >>> lg = lg.to(torch.float64)
      >>> po.tools.remove_grad(lg)
      >>> eig = po.synth.Eigendistortion(img, lg)
      >>> # grab saved example eigendistortion, which runs the above to completion.
      >>> eig.load(
      ...     fetch_data("example_eigendistortion.pt"),
      ...     map_location="cpu",
      ...     tensor_equality_atol=1e-7,
      ... )
      >>> po.synth.eigendistortion.display_eigendistortion_all(eig)
      <PyrFigure size ...>
    """
    # reshape so channel dim is last
    im_shape = eigendistortion._image_shape
    image = eigendistortion.image.detach().view(1, *im_shape).cpu()
    if not hasattr(eigenindex, "__iter__"):
        eigenindex = [eigenindex]
    if not hasattr(alpha, "__iter__"):
        alpha = [alpha] * len(eigenindex)
    if len(alpha) != len(eigenindex):
        raise ValueError(
            "If alpha is a list, it must have the same number of values as eigenindex! "
            f"{len(alpha)=} vs. {len(eigenindex)=}"
        )
    distortions = [torch.ones_like(image)]
    distortion_titles = [""]
    img_titles = ["Original image"]
    dist_suffix = ""
    process_dist = None
    if process_image is None:

        def process_image(x: Tensor) -> Tensor:
            return x

        if as_rgb:

            def process_dist(x: Tensor) -> Tensor:
                return x + 0.5

            warnings.warn(
                "Adding 0.5 to distortion to plot as RGB image, else matplotlib"
                " clipping will result in a strange looking image..."
            )
            dist_suffix = " + 0.5"
    # the only situation in which the following is not true is if process_image
    # was None and as_rgb=True
    if process_dist is None:
        process_dist = process_image

    img_processed = [process_image(image)]

    for a, idx in zip(alpha, eigenindex):
        dist = (
            eigendistortion.eigendistortions[eigendistortion._indexer(idx)]
            .unsqueeze(0)
            .cpu()
        )
        img_processed.append(torch.clamp(process_image(image + a * dist), 0, 1))
        img_titles.append(f"{a} * Eigendistortion[{idx}]")
        distortion_titles.append(f"Eigendistortion[{idx}]{dist_suffix}")
        distortions.append(process_dist(dist))

    fig = imshow(
        distortions + img_processed,
        plot_complex=plot_complex,
        title=distortion_titles + img_titles,
        col_wrap=len(distortions),
        as_rgb=as_rgb,
        **kwargs,
    )
    fig.axes[0].set_visible(False)

    if suptitle_kwargs is None:
        suptitle_kwargs = {}
    va = suptitle_kwargs.pop("verticalalignment", None)
    va = suptitle_kwargs.pop("va", "bottom") if va is None else "bottom"
    y = suptitle_kwargs.pop("y", 1.05)
    fig.suptitle(suptitle, y=y, va=va, **suptitle_kwargs)

    return fig



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import glob
import inspect
import os
import pathlib
import sys
from importlib.metadata import version

import torch

import plenoptic

sys.path.insert(0, os.path.abspath(".."))
sys.path.insert(0, os.path.abspath("./tutorials/"))


# -- Project information -----------------------------------------------------

project = "plenoptic"
copyright = "2019-2025, Plenoptic authors"
author = "Plenoptic authors"

release: str = version("plenoptic")
# this will grab major.minor.patch (excluding any .devN afterwards, which should only
# show up when building locally during development)
version: str = ".".join(release.split(".")[:3])

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "numpydoc",
    "matplotlib.sphinxext.plot_directive",
    "matplotlib.sphinxext.mathmpl",
    "sphinx.ext.autodoc",
    "sphinx_autodoc_typehints",
    "sphinxcontrib.apidoc",
    "sphinx.ext.intersphinx",
    "sphinx_copybutton",
    "sphinx_togglebutton",
    "myst_nb",
    "sphinxcontrib.bibtex",
    "sphinx_design",
    "sphinx.ext.viewcode",
]

add_module_names = False

intersphinx_mapping = {
    "torch": ("https://docs.pytorch.org/docs/stable/", None),
    "matplotlib": ("https://matplotlib.org/stable/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "pyrtools": ("https://pyrtools.readthedocs.io/en/latest", None),
    "python": ("https://docs.python.org/3/", None),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
source_suffix = [".rst", ".md"]

# The master toctree document.
master_doc = "index"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "**.ipynb_checkpoints"]


# Napoleon settings
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = False

numfig = True

# SPHINX CROSS REFERENCES

add_function_parentheses = False

# numpydoc


# find whether the object is part of plenoptic
def is_part_of_plenoptic(obj):
    try:
        return obj.__module__.startswith("plenoptic")
    except AttributeError:
        # then it's a module
        return obj.__name__.startswith("plenoptic")


def is_interesting(obj):
    # find whether the object is a class or a module
    is_right_object = inspect.isclass(obj) or inspect.ismodule(obj)
    return is_right_object and is_part_of_plenoptic(obj)


def find_module(obj):
    members = inspect.getmembers(obj, is_interesting)
    to_return = []
    for name, mem in set(members):
        if inspect.ismodule(mem):
            # go through the modules recursively
            to_return.extend(find_module(mem))
        else:
            # if they inherit torch Module
            if issubclass(mem, torch.nn.Module):
                to_return.append(f"{mem.__module__}.{name}")
    # remove duplicates
    return set(to_return)


# avoid showing all the torch.nn.Module attributes and methods
numpydoc_show_inherited_class_members = {k: False for k in find_module(plenoptic)}

# to avoid this issue https://stackoverflow.com/a/73294408/4659293
numpydoc_class_members_toctree = False

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# these are for the sphinx_rtd_theme
html_theme_options = {
    "display_version": True,
}

# Path for static files (custom stylesheets or JavaScript)
html_static_path = ["_static"]
html_css_files = ["custom.css"]

# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "plenopticdoc"


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        "plenoptic.tex",
        "plenoptic Documentation",
        "Plenoptic authors",
        "manual",
    ),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "plenoptic", "plenoptic Documentation", [author], 1)]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "plenoptic",
        "plenoptic Documentation",
        author,
        "plenoptic",
        "Visual Information Processing",
        "Miscellaneous",
    ),
]


# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ["search.html"]


# -- Extension configuration -------------------------------------------------
#
# APIDOC
apidoc_module_dir = "../src/plenoptic"

# MATPLOTLIB
# because of the examples in the docstrings, want to default to showing source and not
# showing link. in actual doc pages, this needs to be reversed.
plot_include_source = True
plot_html_show_source_link = False

# MYST

myst_enable_extensions = [
    "colon_fence",
    "dollarmath",
    "amsmath",
]

# SPHINXCONTRIB-BIBTEX
bibtex_bibfiles = ["references.bib"]
bibtex_reference_style = "author_year"

# SPHINX COPYBUTTON
copybutton_exclude = ".linenos, .gp"

# MYST_NB

# max time (in secs) per notebook cell. here, we disable this
nb_execution_timeout = -1
# by default, we don't run any of the notebooks (they take too long). users can override
# this behavior using env variables
if run_nb := os.environ.get("RUN_NB"):
    if run_nb == "1" or run_nb == "all":
        nb_execution_excludepatterns = []
        print("Running all notebooks, things will take longer...")
    else:
        all_nbs = glob.glob("tutorials/**/*md", recursive=True)
        all_nbs = [pathlib.Path(n).stem for n in all_nbs]
        run_globs = [f"*{n}*" for n in run_nb.split(",")]
        nb_execution_excludepatterns = [
            f"*{n}*"
            for n in all_nbs
            if not any([glob.fnmatch.fnmatch(n, g) for g in run_globs])
        ]
        print(f"Excluding notebooks: {nb_execution_excludepatterns}")
else:
    nb_execution_excludepatterns = ["*"]
    print("Not running any notebooks, see CONTRIBUTING for details")

nb_execution_mode = os.environ.get("NB_EXECUTION_MODE", "cache")
nb_execution_raise_on_error = True



"""
Simple filters for visual models.

The functions in this module only create tensors, they do not perform convolution.
"""

import torch
from deprecated.sphinx import deprecated
from torch import Tensor

__all__ = ["gaussian1d", "circular_gaussian2d"]


@deprecated(
    "gaussian1d will be removed soon.",  # noqa: E501
    "1.2.0",
)
def gaussian1d(kernel_size: int = 11, std: int | float | Tensor = 1.5) -> Tensor:
    """
    Create normalized 1D Gaussian.

    1d Gaussian of size ``kernel_size``, centered half-way, with variable std
    deviation, and sum of 1.

    Parameters
    ----------
    kernel_size
        Size of Gaussian. Recommended to be odd so that kernel is properly centered.
    std
        Standard deviation of Gaussian.

    Returns
    -------
    filt:
        1d Gaussian with ``Size([kernel_size])``.

    Raises
    ------
    ValueError
        If ``std`` non-scalar.
    ValueError
        If ``std`` non-positive.

    Examples
    --------
    .. plot::

      >>> from plenoptic.simulate import gaussian1d
      >>> from torch.nn.functional import conv1d
      >>> import torch
      >>> import matplotlib.pyplot as plt
      >>> # define a filter
      >>> kernel_size = 21
      >>> filt = gaussian1d(kernel_size=kernel_size, std=2).reshape(1, 1, kernel_size)
      >>> # define a sinusoid + noise
      >>> sin_plus_noise = torch.sin(
      ...     torch.linspace(0, 5 * torch.pi, 500)
      ... ) + 0.5 * torch.randn(500)
      >>> sin_plus_noise = sin_plus_noise.reshape(1, 1, 500)
      >>> # convolve signal with the Gaussian filter
      >>> smooth_sin = conv1d(sin_plus_noise, filt, padding="same")
      >>> # plot filter, signal and convolved signal
      >>> f, axs = plt.subplots(3, 1)
      >>> # plot filter and convolution results
      >>> axs[0].plot(filt.flatten())
      [...
      >>> axs[0].set_title("1D Gaussian filter")
      Text(0.5, 1.0, '1D Gaussian filter')
      >>> axs[1].plot(sin_plus_noise.flatten())
      [...
      >>> axs[1].set_title("Sin + Noise")
      Text(0.5, 1.0, 'Sin + Noise')
      >>> axs[2].plot(smooth_sin.flatten())
      [...
      >>> axs[2].set_title("Convolved Sin + Noise")
      Text(0.5, 1.0, 'Convolved Sin + Noise')
      >>> plt.tight_layout()
    """
    try:
        dtype = std.dtype
    except AttributeError:
        dtype = torch.float32
    std = torch.as_tensor(std, dtype=dtype)
    if std.numel() != 1:
        raise ValueError("std must have only one element!")
    if std <= 0:
        raise ValueError("std must be positive!")
    device = std.device

    x = torch.arange(kernel_size).to(device)
    mu = kernel_size // 2
    gauss = torch.exp(-((x - mu) ** 2) / (2 * std**2))
    filt = gauss / gauss.sum()  # normalize
    return filt


def circular_gaussian2d(
    kernel_size: int | tuple[int, int],
    std: int | list[int] | float | list[float] | Tensor,
    out_channels: int | None = None,
) -> Tensor:
    """
    Create normalized, centered circular 2D gaussian tensor with which to convolve.

    The filter is normalized by total pixel-sum (*not* by ``2*pi*std``) and has shape
    ``(out_channels, 1, height, width)``. For 2d convolutions in torch, the first
    dimensions of the filter tensor corresponds to ``out_channels`` and the second to
    ``in_channels``, see :class:`torch.nn.Conv2d` for more details.

    Parameters
    ----------
    kernel_size
        Filter kernel size. Recommended to be odd so that kernel is properly centered.
        If you use same-padding, convolution with an odd-length kernel will be faster,
        see :func:`torch.nn.functional.conv2d`.
    std
        Standard deviation of 2D circular Gaussian. If a scalar and ``out_channels`` is
        not ``None``, all out channels will have the same value. If not a scalar and
        ``out_channels`` is not ``None``, ``len(std)`` must equal ``out_channels``.
    out_channels
        Number of output channels. If ``None``, inferred from shape of ``std``.

    Returns
    -------
    filt:
        Circular gaussian kernel.

    Raises
    ------
    ValueError:
        If out_channels is not a positive integer.
    ValueError:
        If kernel_size is not one or two positive integers.
    ValueError:
        If std is not positive.
    ValueError:
        If std is non-scalar and ``len(std) != out_channels``

    See Also
    --------
    :class:`~plenoptic.simulate.models.naive.Gaussian`
        Torch Module to perform this convolution.

    Examples
    --------
    Single output channel.

    .. plot::

      >>> import plenoptic as po
      >>> from plenoptic.simulate import circular_gaussian2d
      >>> from torch.nn.functional import conv2d
      >>> import torch
      >>> import matplotlib.pyplot as plt
      >>> kernel_size = 32
      >>> filt_2d = circular_gaussian2d(kernel_size=kernel_size, std=2)
      >>> filt_2d.shape
      torch.Size([1, 1, 32, 32])
      >>> einstein_img = po.data.einstein()
      >>> blurred_einstein = conv2d(einstein_img, filt_2d, padding="same")
      >>> po.imshow(
      ...     [einstein_img, filt_2d, blurred_einstein],
      ...     title=["Einstein", "2D Gaussian Filter", "Blurred Einstein"],
      ... )
      <PyrFigure ...>

    Multiple output channels with different standard deviations.

    .. plot::

      >>> import plenoptic as po
      >>> from plenoptic.simulate import circular_gaussian2d
      >>> from torch.nn.functional import conv2d
      >>> import torch
      >>> import matplotlib.pyplot as plt
      >>> kernel_size = 32
      >>> filt_2d = circular_gaussian2d(
      ...     kernel_size=kernel_size, std=[2, 5.5], out_channels=2
      ... )
      >>> filt_2d.shape
      torch.Size([2, 1, 32, 32])
      >>> einstein_img = po.data.einstein()
      >>> blurred_einstein = conv2d(einstein_img, filt_2d, padding="same")
      >>> titles = [
      ...     "Einstein",
      ...     "2D Gaussian Filter",
      ...     "Larger 2D Gaussian Filter",
      ...     "Blurred Einstein",
      ...     "Blurrier Einstein",
      ... ]
      >>> po.imshow([einstein_img, filt_2d, blurred_einstein], title=titles)
      <PyrFigure ...>

    Multiple input and output channels, convolved independently. See
    :func:`torch.nn.functional.conv2d` to understand the behavior below:

    .. plot::

      >>> import plenoptic as po
      >>> from plenoptic.simulate import circular_gaussian2d
      >>> from torch.nn.functional import conv2d
      >>> import torch
      >>> import matplotlib.pyplot as plt
      >>> kernel_size = 32
      >>> filt_2d = circular_gaussian2d(
      ...     kernel_size=kernel_size, std=[2, 5.5], out_channels=2
      ... ).repeat(3, 1, 1, 1)
      >>> filt_2d.shape
      torch.Size([6, 1, 32, 32])
      >>> wheel = po.data.color_wheel(as_gray=False)
      >>> blurred_wheel = conv2d(wheel, filt_2d, groups=3, padding="same")
      >>> titles = ["Wheel", "Blurred Wheel", "Blurrier Wheel"]
      >>> # note that the order of channels: the first two correspond to the first
      >>> # channel of the input image, convolved with the each of the two gaussians,
      >>> # and so on.
      >>> po.imshow(
      ...     [wheel, blurred_wheel[:, ::2], blurred_wheel[:, 1::2]],
      ...     title=titles,
      ...     as_rgb=True,
      ... )
      <PyrFigure ...>
    """
    kernel_size, std, out_channels = _validate_filter_args(
        kernel_size, std, out_channels
    )

    origin = (kernel_size + 1) / 2

    shift_y = torch.arange(1, kernel_size[0] + 1, device=std.device) - origin[0]
    shift_x = torch.arange(1, kernel_size[1] + 1, device=std.device) - origin[1]

    (xramp, yramp) = torch.meshgrid(shift_x, shift_y, indexing="xy")

    log_filt = (xramp**2) + (yramp**2)
    log_filt = log_filt.repeat(out_channels, 1, 1, 1)
    log_filt = log_filt / (-2.0 * std**2).view(out_channels, 1, 1, 1)

    filt = torch.exp(log_filt)
    # normalize
    filt = filt / torch.sum(filt, dim=[1, 2, 3], keepdim=True)

    return filt


def _validate_filter_args(
    kernel_size: int | tuple[int, int],
    std: int | list[int] | float | list[float] | Tensor,
    out_channels: int | None,
    std_name: str = "std",
    out_channels_name: str = "out_channels",
) -> tuple[Tensor, Tensor, Tensor]:
    """
    Validate common filter args.

    Checks that:

    - kernel_size is positive, integer-valued, and has 1 or 2 values

    - std is positive and either a single value (i.e., an int, float, or scalar tensor)
      or ``len(std) == out_channels``

    - out_channels must be a positive integer.

    Does the following and then returns the three values

    - if ``out_channels`` is ``None``, then infer from shape of ``std``.

    - makes ``kernel_size`` a 1d tensor of size 2

    - makes ``std`` a float32 1d tensor of size ``out_channels``

    Parameters
    ----------
    kernel_size
        Filter kernel size.
    std
        Standard deviation of 2D circular Gaussian. If a scalar and ``out_channels`` is
        not ``None``, all out channels will have the same value. If not a scalar and
        ``out_channels`` is not ``None``, ``len(std)`` must equal ``out_channels``.
    out_channels
        Number of output channels. If ``None``, inferred from ``len(std)``.
    std_name, out_channels_name
        Names of these variables to raise more informative error messages (when e.g.,
        calling from ``CenterSurround``, which uses this function to validate different
        std arguments).

    Returns
    -------
    kernel_size, std, out_channels
        The validated tensors.

    Raises
    ------
    ValueError
        If out_channels is not a positive integer.
    ValueError
        If kernel_size is not one or two positive integers.
    ValueError
        If std is not positive.
    ValueError
        If std is non-scalar and ``len(std) != out_channels``
    """
    std = torch.as_tensor(std)
    if not torch.is_floating_point(std):
        std = std.to(torch.float32)

    if out_channels is None:
        out_channels = len(std) if std.ndim != 0 else 1

    if out_channels < 1 or isinstance(out_channels, float):
        raise ValueError(f"{out_channels_name} must be positive integer")

    if std.ndim == 0:
        std = std.repeat(out_channels)

    kernel_size = torch.as_tensor(kernel_size).to(std.device)
    if kernel_size.numel() == 1:
        kernel_size = kernel_size.repeat(2)
    if torch.is_floating_point(kernel_size):
        raise ValueError("kernel_size must be integer-valued")
    if torch.any(kernel_size < 1):
        raise ValueError("kernel_size must be positive")

    if torch.any(std <= 0):
        raise ValueError(f"{std_name} must be positive")
    if len(std) != out_channels:
        raise ValueError(
            f"If non-scalar, len({std_name}) must equal {out_channels_name}"
        )

    return kernel_size, std, out_channels



"""
Naive image metrics.

These aren't expected to do very well, just to provide a baseline for comparison.
"""

import torch


def mse(img1: torch.Tensor, img2: torch.Tensor) -> torch.Tensor:
    r"""
    Return the MSE between img1 and img2.

    Our baseline metric to compare two images is often mean-squared
    error, MSE. This is not a good approximation of the human visual
    system, but is handy to compare against.

    For two images, :math:`x` and :math:`y`, with :math:`n` pixels
    each:

    .. math::

        MSE = \frac{1}{n}\sum_i (x_i - y_i)^2

    Parameters
    ----------
    img1
        The first image to compare.
    img2
        The second image to compare, must be same size as ``img1``.

    Returns
    -------
    mse
        The mean-squared error between ``img1`` and ``img2``.

    Raises
    ------
    RuntimeError
        If ``img1`` and ``img2`` aren't the same size.

    Examples
    --------
    >>> import plenoptic as po
    >>> import torch
    >>> einstein = po.data.einstein()
    >>> po.tools.set_seed(0)
    >>> einstein_noisy = einstein + 0.1 * torch.randn_like(einstein)
    >>> po.metric.mse(einstein, einstein_noisy)
    tensor([[0.0100]])
    """
    return torch.pow(img1 - img2, 2).mean((-1, -2))
